/*
 * mmapio.c
 * Copyright (C) 1999 by Christopher Nelson
 *
 * Some changes by Richard Dawe (RD) <richdawe@bigfoot.com>, when writing some
 * share memory functions, to provide an ioctl to retrieve the memory buffer
 * offset, which is used in the shmat() and shmdt() functions.
 */

#include <sys/fsext.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <dir.h>
#include <unistd.h>
#include <io.h>
#include <errno.h> /* RD: Need errno & codes */
#include "mmio.h"

static MMFILE **__mmio_list=0;
static unsigned int n_mmio_handles=0;
static int mmio_handler(__FSEXT_Fnumber func_number, int *rv, va_list args);

#define BLOCK_SIZE (1<<14)

static int
init()
{
 /* if there's anything we need to do, we should do it here. */
 __FSEXT_add_open_handler(mmio_handler);
 return(0);
}

static int
grow_table(unsigned int size)
{

 if (size<n_mmio_handles)
   return(0);
 else
   {
    unsigned int i;

    __mmio_list = (MMFILE **)realloc(__mmio_list, size*sizeof(MMFILE *));
    if (__mmio_list==0)
      return(1);

    for(i=n_mmio_handles; i<size; i++)
	__mmio_list[i]=0;
   }

 n_mmio_handles=size;
 return(0);
}

static int
grow_file(int fd, unsigned long how_much)
{
 MMFILE *mmf =  (MMFILE *)__FSEXT_get_data(fd);
 mmf->buffer = realloc(mmf->buffer, mmf->size+how_much);
 if (mmf->buffer==0)
   return(1);

 mmf->size+=how_much;
 return(0);
}

static char *
mmf_get_filename(char *path)
{
 char d[MAXDRIVE], p[MAXDIR], f[MAXFILE], e[MAXEXT], b[MAXPATH], *rv;
 fnsplit(path, d, p, f, e);
 fnmerge(b,0,0,f,e);

 rv = (char *)malloc(strlen(b)+1);
 strcpy(rv, b);
 return(rv);
}

static int
mmf_add(int fd, char *filename)
{
 unsigned int i;
 time_t now;
 MMFILE *mmf;

 mmf = (MMFILE *)malloc(sizeof(MMFILE));
 if (mmf==0)
   return(1);

 memset(mmf, 0, sizeof(MMFILE));
 mmf->filename=filename;
 mmf->buffer=malloc(BLOCK_SIZE);

 if (mmf->buffer==0)
   {
    free(mmf);
    return(1);
   }

 time(&now);
 mmf->atime=now;
 mmf->ctime=now;
 mmf->mtime=now;
 mmf->size = BLOCK_SIZE;

 __FSEXT_set_data(fd, (void *)mmf);

while(1)
  {
 for(i=0; i<n_mmio_handles; i++)
    {
     if (__mmio_list[i]==0)
       {
        __mmio_list[i]=mmf;
	return(0);
       }
    }
   if (grow_table(n_mmio_handles+1))
     return(1);
  }
 return(0);
}

static int
mmf_find_file(char *filename)
{
 unsigned int i;

 for(i=0; i<n_mmio_handles; i++)
    {
     if (__mmio_list[i])
       if (strcmp(__mmio_list[i]->filename, filename)==0)
	 return(i);
    }

 return(-1);
}

static int
mmf_is_valid(char *filename)
{
 char d[MAXDRIVE], p[MAXDIR], f[MAXFILE], e[MAXEXT];
 fnsplit(filename, d, p, f, e);

 if (strcmp(p, "/dev/ramdrv/")==0)
   return(1);
 else
   return(0);
}

int
mmf_creat(char *name, int attr)
{
 char *filename;
 MMFILE *mmf;
 int fd, loc;

 if (!mmf_is_valid(name))
   return(-2);

 filename =  mmf_get_filename(name);
 fd = __FSEXT_alloc_fd(mmio_handler);
 if ((loc=mmf_find_file(filename))>=0)
   {
    free(filename);
    mmf = __mmio_list[loc];

    mmf->position=0;
    mmf->eof=0;
    time(&mmf->atime);
    time(&mmf->ctime);
    time(&mmf->mtime);

    __FSEXT_set_data(fd, (void *)mmf);
   }
 else
   {
    if (mmf_add(fd, filename))
      return(-1);
   }

 return(fd);
}

int
mmf_open(char *name, int attr)
{
 char *filename;
 MMFILE *mmf;
 int fd, loc;

 if (!mmf_is_valid(name))
   return(-2);

 filename = mmf_get_filename(name);

 fd = __FSEXT_alloc_fd(mmio_handler);
 if ((loc=mmf_find_file(filename))>=0)
   {
    free(filename);
    mmf = __mmio_list[loc];

    if (!(attr & O_APPEND)) mmf->position=0;
    if (attr & O_TRUNC)     mmf->eof=0;

    time(&mmf->atime);
    time(&mmf->mtime);

    __FSEXT_set_data(fd, (void *)mmf);
   }
 else if (attr & O_CREAT)
   {
    if (mmf_add(fd, filename))
      return(-1);
   }
 else
   free(filename);

 return(fd);
}

int
mmf_close(int filedes)
{
 __FSEXT_set_data(filedes, 0);
 __FSEXT_set_function(filedes, 0);
 _close(filedes);
 return(0);
}

inline int
mmf_read(int filedes, void *buf, size_t nbyte)
{
    MMFILE *mmf =  (MMFILE*)__FSEXT_get_data(filedes);
    size_t nleft = mmf->eof - mmf->position;

    if (nleft<nbyte)
      nbyte=nleft;

    memcpy(buf, (void *)((char *)mmf->buffer+mmf->position), nbyte);
    mmf->position+=nbyte;

    time(&mmf->atime);
    return(nbyte);
}

inline int
mmf_write(int filedes, void *buf, size_t nbyte)
{
    MMFILE *mmf =  (MMFILE*)__FSEXT_get_data(filedes);
    size_t nleft = mmf->size - mmf->position;

    if (nleft<nbyte)
      if (grow_file(filedes, (nleft > BLOCK_SIZE ? nleft : BLOCK_SIZE)))
	  return(0);

    memcpy((void *)((char *)mmf->buffer+mmf->position), buf, nbyte);
    mmf->position+=nbyte;
    mmf->eof = (mmf->eof < mmf->position ? mmf->position : mmf->eof);

    time(&mmf->ctime);
    return(nbyte);
}

inline off_t
mmf_lseek(int filedes, off_t offset, int whence)
{
    MMFILE *mmf =  (MMFILE*)__FSEXT_get_data(filedes);
    off_t  final_pos;

    switch(whence)
      {
       case SEEK_SET:
	    final_pos=offset;
	    break;

       case SEEK_CUR:
	    final_pos=(long)mmf->position+offset;
	    break;

       case SEEK_END:
	    final_pos=(long)mmf->eof+offset;
	    break;

       default:
            final_pos=offset;
            break;
      }

    mmf->position = (final_pos >=0 ?
		     (final_pos <= (long)mmf->eof ?
		      final_pos
		      :	mmf->eof)
		     : 0 );

    return(mmf->position);
}

int
mmf_fstat(int filedes, struct stat *sbuf)
{
    MMFILE *mmf =  (MMFILE*)__FSEXT_get_data(filedes);

    sbuf->st_dev     = -1;
    sbuf->st_atime   = mmf->atime;
    sbuf->st_ctime   = mmf->ctime;
    sbuf->st_ino     = (ino_t)(mmf->buffer)<<16;
    sbuf->st_mode    = S_IRUSR | S_IWUSR;
    sbuf->st_mtime   = mmf->mtime;
    sbuf->st_size    = mmf->eof;
    sbuf->st_nlink   = 1;
    sbuf->st_blksize = BLOCK_SIZE;
    sbuf->st_gid     = getgid();
    sbuf->st_uid     = getuid();

    return(0);
}

int mmf_stat(char *filename, struct stat *sbuf)
{
 int filedes;
 char *name;

 if (!mmf_is_valid(filename))
   return(-2);

 name = mmf_get_filename(filename);
 if ((filedes = mmf_find_file(name))>=0)
   {
    MMFILE *mmf =  __mmio_list[filedes];
 
    sbuf->st_dev     = -1;
    sbuf->st_atime   = mmf->atime;
    sbuf->st_ctime   = mmf->ctime;
    sbuf->st_ino     = (ino_t)(mmf->buffer)<<16;
    sbuf->st_mode    = S_IRUSR | S_IWUSR;
    sbuf->st_mtime   = mmf->mtime;
    sbuf->st_size    = mmf->eof;
    sbuf->st_nlink   = 1;
    sbuf->st_blksize = BLOCK_SIZE;
    sbuf->st_gid     = getgid();
    sbuf->st_uid     = getuid();

    free(name);
    return(0);
   }
 else
   {
    free(name);
    return(-1);
   }
}

int mmf_unlink(char *filename)
{
 int filedes;
 char *name;

 if (!mmf_is_valid(filename))
   return(-2);

 name = mmf_get_filename(filename);
 if ((filedes = mmf_find_file(name))>=0)
   {
    MMFILE *mmf =  __mmio_list[filedes];

    free(name);
    free(mmf->filename);
    free(mmf->buffer);
    free(mmf);

    __mmio_list[filedes]=0;
    return(0);
   }
 else
   {
    free(name);
    return(1);
   }
}

static int
mmio_handler(__FSEXT_Fnumber func_number, int *rv, va_list args)
{
 int fd, whence, attr;
 void *buf;
 char *n;
 size_t size;
 off_t  offset;
 struct stat *ss;

 /* RD: For ioctl() support */
 int cmd;
 void **ioctl_arg;
 MMFILE *mmf;
 /* End RD */

 switch(func_number)
   {
  case   __FSEXT_nop:	return(1); break; /* does nothing: nop */
  case   __FSEXT_open:
       n = va_arg(args, char *);
       attr = va_arg(args, int);
       *rv=mmf_open(n, attr);
       return(*rv==-2 ? 0 : 1);
       break;

  case   __FSEXT_creat:
       n = va_arg(args, char *);
       attr = va_arg(args, int);
       *rv=mmf_creat(n, attr);
       return(*rv==-2 ? 0 : 1);
       break;

  case   __FSEXT_read:
       fd = va_arg(args, int);
       buf = va_arg(args, void *);
       size = va_arg(args, size_t);
       *rv = mmf_read(fd, buf, size);
       return(1);
       break;

  case   __FSEXT_write:
       fd = va_arg(args, int);
       buf = va_arg(args, void *);
       size = va_arg(args, size_t);
       *rv = mmf_write(fd, buf, size);
       return(1);
       break;

  case   __FSEXT_ready:
       /* RD: The return value should be returned in rv. */
       /*return(__FSEXT_ready_read | __FSEXT_ready_write);*/
       *rv = __FSEXT_ready_read | __FSEXT_ready_write;
       return(1);
       break;

  case   __FSEXT_close:
       *rv = mmf_close(va_arg(args, int));
       return(1);
       break;

  case   __FSEXT_fcntl:
       *rv=-1;		/* FIXME: what do the commands do? */
       return(1);

  case   __FSEXT_ioctl:
	/* RD: Added support for MMIOGETBUFADDR ioctl, to get the buffer
	 * address of an mmapio file. */
	fd  = va_arg(args, int);
	cmd = va_arg(args, int);

	switch(cmd) {
		case MMIOGETBUFADDR:
			ioctl_arg  = va_arg(args, void **);
			mmf        = (MMFILE *)__FSEXT_get_data(fd);
			*ioctl_arg = mmf->buffer;
			*rv = 0;
			break;

		default:
			/* Unsupported ioctl */
			errno = EINVAL;
			*rv   = -1;
			break;
	}
	/* End RD */

       /**rv = 0;*/		/* FIXME: is this really neccessary? */
       return(1);

  case   __FSEXT_lseek:
       fd = va_arg(args, int);
       offset =	va_arg(args, off_t);
       whence =	va_arg(args, int);
       *rv = mmf_lseek(fd, offset, whence);
       return(1);

  case   __FSEXT_link:
       *rv = 1;		/* FIXME: actually make this work.  *link* them. */
       return(1);

  case   __FSEXT_unlink:
       *rv = mmf_unlink(va_arg(args, char *));
       return(*rv==-2 ? 0 : 1);

  case   __FSEXT_dup:
       *rv = -1;
       return(1);

  case   __FSEXT_dup2:
       *rv = -1;
       return(1);

  case   __FSEXT_fstat:
       fd = va_arg(args, int);
       ss = va_arg(args, struct stat *);
       *rv = mmf_fstat(fd, ss);
       return(1);

  case   __FSEXT_stat:
       n = va_arg(args, char *);
       ss = va_arg(args, struct stat *);
       *rv = mmf_stat(n, ss);
       return(*rv==-2 ? 0 : 1);

  default:
       return(0);
  }
}

void __attribute__((constructor))
mmf_constructor()
{
 init();
}

void __attribute__((destructor))
mmf_destructor()
{
 unsigned int i;

 for(i=0; i<n_mmio_handles; i++)
    {
     if (__mmio_list[i])
       {
	free(__mmio_list[i]->filename);
	free(__mmio_list[i]->buffer);
	free(__mmio_list[i]);
       }
    }
}

