/*
   shm.c - Shared memory functions
   Written by Richard Dawe <richdawe@bigfoot.com>

   This makes use of Christopher Nelson's /dev/ramdrv DJGPP File System
   Extension - see the files 'mmapio.c' & 'mmio.h' for more details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
/*#include <sys/types.h>
#include <sys/mman.h>*/

#include "mmio.h"
#include <sys/ipc.h>
#include <sys/shm.h>

typedef struct _shm_info {
	struct shmid_ds desc; /* Shared memory info            */
	key_t           key;  /* Key from ftok()               */
	char           *name; /* File name for mmio            */
	int             fd;   /* File descriptor               */
	void           *buf;  /* Shared memory buffer          */
} SHM_INFO;

#define SHM_NAME_PREFIX "/dev/ramdrv/"

/* shmid is the index to this array */
#define SHM_INFO_MAX 256
static SHM_INFO shm_info[SHM_INFO_MAX];

#define SHM_OPEN_DEFAULT_SIZE 65536

/* ------------
 * - shm_init -
 * ------------ */

static int shm_init (void)
{
	static int shm_inited = 0;
	int i;

	if (shm_inited) return(shm_inited);

	/* Clear info structs */
	for (i = 0; i < SHM_INFO_MAX; i++) {
		bzero(&shm_info[i], sizeof(SHM_INFO));
	}

	/* Set up the memory-mapped I/O handler */
	mmf_constructor();

	shm_inited = 1;
	return(shm_inited);
}

/* shm_init() should be called automatically. */
void __attribute__((constructor)) shm_constructor (void) { shm_init(); }

/* TODO: atexit() handler? */

/* ---------------
 * - shmid_valid -
 * --------------- */

/* Check that the specified shmid is valid; returns: 1 = yes, 0 = no. */

static int shmid_valid (int shmid)
{
	if (shmid >= SHM_INFO_MAX) return(0);
	if (shm_info[shmid].key == 0) return(0);
	return(1);
}

/* ---------
 * - shmat -
 * --------- */

/* NOTE: This ignores the Unix98 spec's details for shmaddr and returns the
 * address of the memory-mapped file. According to the Unix98 spec, it should
 * return shmaddr (or a pointer to a nearby area) if shmaddr != NULL. */

/* TODO: I wonder if the spec could be implemented using DPMI magic, i.e.
 * map the buffer to multiple points in memory? */

void *shmat (int shmid, const void *shmaddr, int shmflag)
{
	/* No such piece of shared memory! */
	if (!shmid_valid(shmid)) {
		errno = EINVAL;
		return((void *) -1);
	}

	/*if ((shmflg & SHM_RDONLY) == SHM_RDONLY) {
	}*/

	/* Update info struct */
	time(&shm_info[shmid].desc.shm_atime);
	shm_info[shmid].desc.shm_nattch++;
	return(shm_info[shmid].buf);
}

/* ----------
 * - shmctl -
 * ---------- */

int shmctl (int shmid, int cmd, struct shmid_ds *buf)
{
	/* No such piece of shared memory! */
	if (!shmid_valid(shmid)) {
		errno = EINVAL;
		return(-1);
	}

	switch(cmd) {
		/* Get shm ifo structure */
		case IPC_STAT:
			/* Bad buffer */
			if (buf == NULL) {
				errno = EFAULT;
				return(-1);
			}

			memcpy(buf, &shm_info[shmid].desc,
			       sizeof(shm_info[shmid].desc));
			break;

		/* Update permissions for shm */
		case IPC_SET:
			/* Bad buffer */
			if (buf == NULL) {
				errno = EFAULT;
				return(-1);
			}

			/* Unlikely with DJGPP ;) */
			if ((shm_info[shmid].desc.shm_perm.cuid != geteuid())
			    && (shm_info[shmid].desc.shm_perm.uid != geteuid())
			   ) {
				errno = EPERM;
				return(-1);
			}

			/* TODO: This should really change the file access mode
			 * too. */
			shm_info[shmid].desc.shm_perm.uid  = buf->shm_perm.uid;
			shm_info[shmid].desc.shm_perm.gid  = buf->shm_perm.gid;
			shm_info[shmid].desc.shm_perm.mode = buf->shm_perm.mode;

			/* Update info struct with time info */
			time(&shm_info[shmid].desc.shm_ctime);

			break;

		case IPC_RMID:
			/* Unlikely with DJGPP ;) */
			if ((shm_info[shmid].desc.shm_perm.cuid != geteuid())
			    && (shm_info[shmid].desc.shm_perm.uid != geteuid())
			   ) {
				errno = EPERM;
				return(-1);
			}

			/* Destroy the shared memory */
			close(shm_info[shmid].fd);
			free(shm_info[shmid].name);
			bzero(&shm_info[shmid], sizeof(SHM_INFO));
			break;

		default:
			errno = EINVAL;
			return(-1);
	}

	return(0);
}

/* ---------
 * - shmdt -
 * --------- */

int shmdt (const void *shmaddr)
{
	int shmid;
	int i;

	/* Find the shmid that this corresponds to. */
	for (shmid = -1, i = 0; i < SHM_INFO_MAX; i++) {
		if (shm_info[i].key == 0) continue;

		if (shm_info[i].buf == shmaddr) {
			shmid = i;
			break;
		}
	}

	/* Not found! */
	if (shmid == -1) {
		errno = EINVAL;
		return(-1);
	}

	/* Update info struct */
	time(&shm_info[shmid].desc.shm_dtime);

	/* Decrement attachment counter */
	shm_info[shmid].desc.shm_nattch--;

	return(0);
}

/* ----------
 * - shmget -
 * ---------- */

int shmget (key_t key, size_t size, int shmflg)
{
	int i, shmid, fd, ret;
	char name[PATH_MAX];
	char buf[1024];

	/* Memory protection vars */
	/*int   pflags, poffset;
	int   psize;
	void *pbuf;*/

	/* SHM exists? */
	for (i = 0; i < SHM_INFO_MAX; i++) {
		if (key == IPC_PRIVATE)     break;
		if (key != shm_info[i].key) continue;

		/* Found it, check shmget() parameters */
		if ((shmflg & IPC_CREAT) && (shmflg & IPC_EXCL)) {
			errno = EEXIST;
			return(-1);
		}

		/* NB: Array index = shmid */
		return(i);
	}

	/* Key not found & creation not specified */
	if (!(shmflg & IPC_CREAT) && (key != IPC_PRIVATE)) {
		errno = ENOENT;
		return(-1);
	}

	/* Find a free id */
	for (shmid = -1, i = 0; i < SHM_INFO_MAX; i++) {
		if (shm_info[i].key == 0) {
			shmid = i;
			break;
		}
	}

	if (shmid == -1) {
		/* No space! */
		errno = ENOSPC;
		return(-1);
	}

	/* TODO: Upper bound on size of shared memory segment? */

	/* Build the info structure */
	sprintf(name, "%s%04x.shm", SHM_NAME_PREFIX, shmid);
	fd = open(name, O_RDWR | O_CREAT | O_BINARY, S_IRUSR | S_IWUSR);
	if (fd < 0) {
		errno = ENOMEM; /* TODO: Right error? */
		return(-1);
	}

	bzero(&shm_info[shmid], sizeof(SHM_INFO));
	shm_info[shmid].desc.shm_perm.cuid = geteuid();
	shm_info[shmid].desc.shm_perm.uid  = geteuid();
	shm_info[shmid].desc.shm_perm.cgid = getegid();
	shm_info[shmid].desc.shm_perm.gid  = getegid();
	shm_info[shmid].desc.shm_perm.mode = shmflg && 0x1FF;
	shm_info[shmid].desc.shm_segsz     = size;
	shm_info[shmid].desc.shm_cpid      = getpid();
	shm_info[shmid].key                = key;
	shm_info[shmid].name               = strdup(name);
	shm_info[shmid].fd                 = fd;

	/* Blank the shared memory segment & expand the I/O buffer to
	 * accomodate it. */
	lseek(fd, SEEK_SET, 0);

#define SHMGET_CLEANUP						\
	{							\
		close(shm_info[shmid].fd);			\
		free(shm_info[shmid].name);			\
		bzero(&shm_info[shmid], sizeof(SHM_INFO));	\
		errno = ENOMEM;					\
		return(-1);					\
	}

	bzero(buf, sizeof(buf));
	for (i = 0; i < (size / sizeof(buf)); i++) {
		ret = write(fd, buf, sizeof(buf));
		if (ret == -1) SHMGET_CLEANUP;
	}
	ret = write(fd, buf, size % sizeof(buf));
	if (ret == -1) SHMGET_CLEANUP;

	/* Store the buffer's address for shmat(), shmdt() functions */
	ret = ioctl(fd, MMIOGETBUFADDR, &shm_info[shmid].buf);
	if (ret == -1) SHMGET_CLEANUP;

	/* Protect the memory buffer according to shmflg */
	/*pflags = PROT_NONE;
	if ((shmflg & S_IRUSR) == S_IRUSR) pflags |= PROT_READ;
	if ((shmflg & S_IWUSR) == S_IWUSR) pflags |= PROT_WRITE;*/

	/* TODO: This really requires the memory buffer to be 4K-aligned, so
	 * that we don't protect parts that aren't ours, or leave parts
	 * unprotected. So leave it for now. */

	/* TODO: This needs testing/checking with a DPMI 1.0 client
	 * (or CWSDPMI?). */
	/* 4K page size = minimum protectable area, also 4K aligned */
#define ALIGNED_4K_OFFSET(x) ((((int) x / 4096) * 4096) - (int) x)
	/*poffset = ALIGNED_4K_OFFSET(shm_info[shmid].buf);
	pbuf    = (void *) ((int) shm_info[shmid].buf + poffset);
	psize   = (int) size - poffset;
	psize   = (psize <= 0) ? 0 : (psize / 4096) * 4096;

	if (psize > 0) mprotect(pbuf, (size_t) psize, pflags);*/

	return(shmid);
}
